/* UFO Extender - Mod for X-COM
 * See <http://www.ufopaedia.org/index.php?title=User:Seb76> for more details.
 * Copyright (C) 2008 Seb76
 *
 * This file is part of UFO Extender.

 * UFO Extender is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * UFO Extender is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with UFO Extender.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "stdafx.h"
#include "base.h"
#include "xcom.h"
#include "patcher.h"

/*
 * Called when a facility is destroyed. We update the radar stats to fix the phantom radar bug
 */
void PhantomRadarFix()
{
	UpdateRadarStats(*currentBaseIdx);
}


/*
 * Fix for the base disjoint bug
 * We scan the border modules and fix the tiles when needed
 */
void BaseDisjointFix()
{
	((void (*)())(0x41BF90))();

	unsigned char *modules=&(pGeodata[12]);
	int *groundTiles=&((*pMap)[(*map_numTilesPerLevel)]);

#define MODULE(X,Y) modules[(X)+6*(Y)]
#define TILE(X,Y) groundTiles[(X)+60*(Y)]

	//Fix modules at the bottom
	for(int i=0;i<5;i++)
	{
		int j=5;
		{

			if(MODULE(i,j)>=(unsigned char)0x14)
				continue;
			if(MODULE(i,j)==0x12) //bottom left hangar
				continue;
			if(MODULE(i+1,j)>=(unsigned char)0x14)
				continue;

			char *tile1=(char *)&TILE(i*10+9,j*10+3);
			char *tile2=(char *)&TILE(i*10+9,j*10+4);
			char *tile3=(char *)&TILE(i*10+9,j*10+5);

			char *tile4=(char *)&TILE(i*10+10,j*10+3);
			char *tile5=(char *)&TILE(i*10+10,j*10+4);
			char *tile6=(char *)&TILE(i*10+10,j*10+5);

			char *tile7=(char *)&TILE(i*10+9,j*10+6);

			tile1[0]=9;
			tile1[1]=0;
			tile1[2]=16;
			tile1[3]=0;

			tile2[0]=9;
			tile2[1]=0;
			tile2[2]=0;
			tile2[3]=0;

			tile3[0]=9;
			tile3[1]=0;
			tile3[2]=0;
			tile3[3]=0;

			//Remove west wall
			tile4[1]=0;
			tile5[1]=0;
			tile6[1]=0;

			tile7[2]=16;
		}
	}
	//Fix modules on the right edge
	int i=5;
	{
		for(int j=0;j<5;j++)
		{
			if(MODULE(i,j)>=(unsigned char)0x14)
				continue;
			if(MODULE(i,j)==0x11) //top right hangar
				continue;
			if(MODULE(i,j+1)>=(unsigned char)0x14)
				continue;

			char *tile1=(char *)&TILE(i*10+3,j*10+9);
			char *tile2=(char *)&TILE(i*10+4,j*10+9);
			char *tile3=(char *)&TILE(i*10+5,j*10+9);

			char *tile4=(char *)&TILE(i*10+3,j*10+10);
			char *tile5=(char *)&TILE(i*10+4,j*10+10);
			char *tile6=(char *)&TILE(i*10+5,j*10+10);

			char *tile7=(char *)&TILE(i*10+6,j*10+9);

			tile1[0]=9;
			tile1[1]=15;
			tile1[2]=0;
			tile1[3]=0;

			tile2[0]=9;
			tile2[1]=0;
			tile2[2]=0;
			tile2[3]=0;

			tile3[0]=9;
			tile3[1]=0;
			tile3[2]=0;
			tile3[3]=0;

			//Remove north wall
			tile4[2]=0;
			tile5[2]=0;
			tile6[2]=0;

			tile7[1]=15;
		}
	}
}

/*
 * We replace the standard radar coverage calculation by this one which
 * takes radar stacking into account
 */
void FixRadarStacking(__int16 baseIdx)
{
	unsigned char *base=(*pBaseList)+baseIdx*292;
	unsigned char *modules=&base[22];
	unsigned char *completion=&base[58];
	int smallRadars=0;
	int largeRadars=0;
	int hwd=0;

	//Scan base modules
	for(int i=0;i<6;i++)
	{
		for(int j=0;j<6;j++)
		{
			if(*completion==0)
			{
				switch(*modules)
				{
				case 4://small radar
					smallRadars++;
					break;
				case 5://large radar
					largeRadars++;
					break;
				case 15://hyper wave decoder
					hwd++;
					break;
				}
			}
			modules++;
			completion++;
		}
	}

	float shortDetection=pow(0.9f,smallRadars);
	float largeDetection=pow(0.8f,largeRadars);

	*(short *)(&base[0x10])=(short)((1.0-shortDetection*largeDetection)*100.0);
	*(short *)(&base[0x12])=(short)((1.0-largeDetection)*100.0);
	*(short *)(&base[0x14])=hwd?100:0;
}

/*
 * Queued construction
 */
__declspec(naked) void PatchBaseConstruction()
{
	__asm
	{
		test cx,cx
		jnz can_build

//0x470644
		mov     [eax+16h], dl
		mov     dl, 0xff
		mov     [eax+3Ah], dl
		mov     ecx,1
		push 0x433E36
		//push 0x433E2D //cannot build here msg
		ret
can_build:
		push 0x433E12
		ret
	}
}


__declspec(naked) void PatchBaseConstructionHangar()
{
	__asm
	{
		test si,si
		jnz can_build

//0x470644
		mov     [eax+16h], 0x10
		mov     [eax+17h], 0x11
		mov     [eax+1Ch], 0x12
		mov     [eax+1Dh], 0x13
		mov     dl, 0xff
		mov     [eax+3ah], dl
		mov     [eax+3bh], dl
		mov     [eax+40h], dl
		mov     [eax+41h], dl
		mov     ecx,1
		push 0x433E36
		//push 0x433E2D //cannot build here msg
		ret
can_build:
		push 0x433CF8
		ret
	}
}
__declspec(naked) void PatchValidSquares()
{
	__asm
	{
		mov ebp,[esp+0x14] //&base[0x16]
		mov eax,0x049C2C0 //valid square list (word *)

		mov edi,6*6

squares:
		cmp byte ptr [ebp], 0xff
		jz skip
		cmp byte ptr [ebp+0x24], 0xff
		jnz skip
		mov word ptr [eax],1
skip:
		add eax,2
		inc ebp
		dec edi
		jnz squares


		//tail of the original function
		pop     edi
		pop     esi
		pop     ebp
		pop     ebx
		add     esp, 8
		ret
	}
}

void BaseDrawRemainingDays(int number, int color, __int16 ypos, __int16 xpos, int useSmallFonts)
{
	if(number!=0xff)
	{
		if(number<10)
			xpos+=4;
		DrawNumber2(number, color, ypos, xpos, useSmallFonts);
	}
	else
	{
		DrawString("Queued",color,ypos,xpos-5,1);
	}
}

static unsigned int _edi;

void BaseFacilityCompleted(int baseIdx)
{
	unsigned char *base=(*pBaseList)+baseIdx*292;
	unsigned char *modules=&base[22];
	unsigned char *completion=&base[58];

#define MODULE(x,y) modules[(x)+6*(y)]
#define COMPLETION(x,y) completion[(x)+6*(y)]

	for(int i=0;i<6;i++)
	{
		for(int j=0;j<6;j++)
		{
			if(MODULE(i,j)!=0xff && COMPLETION(i,j)==0xff)
			{
				int unqueue=0;

				if(i!=0 && MODULE(i-1,j)!=0xff && COMPLETION(i-1,j)==0)
					unqueue=1;
				else if(i!=5 && MODULE(i+1,j)!=0xff && COMPLETION(i+1,j)==0)
					unqueue=1;
				else if(j!=0 && MODULE(i,j-1)!=0xff && COMPLETION(i,j-1)==0)
					unqueue=1;
				else if(j!=5 && MODULE(i,j+1)!=0xff && COMPLETION(i,j+1)==0)
					unqueue=1;

				if(unqueue==1)
				{
					//if the newly started construction is _after_ the current one
					//add one to its construction time 'cause it'll be decremented
					//right away

					#define DELTA(i,j) ((((unsigned int)&COMPLETION(i,j))>_edi)?1:0)

					unsigned char *constructionData=(unsigned char *)0x470644;
					switch(MODULE(i,j))
					{
						case 0x10:
							COMPLETION(i,j)=25+DELTA(i,j);
							COMPLETION(i+1,j)=25+DELTA(i+1,j);
							COMPLETION(i,j+1)=25+DELTA(i,j+1);
							COMPLETION(i+1,j+1)=25+DELTA(i+1,j+1);
							break;
						case 0x11:
							COMPLETION(i-1,j)=25+DELTA(i-1,j);
							COMPLETION(i,j)=25+DELTA(i,j);
							COMPLETION(i-1,j+1)=25+DELTA(i-1,j+1);
							COMPLETION(i,j+1)=25+DELTA(i,j+1);
							break;
						case 0x12:
							COMPLETION(i,j-1)=25+DELTA(i,j-1);
							COMPLETION(i+1,j-1)=25+DELTA(i+1,j-1);
							COMPLETION(i,j)=25+DELTA(i,j);
							COMPLETION(i+1,j)=25+DELTA(i+1,j);
							break;
						case 0x13:
							COMPLETION(i-1,j-1)=25+DELTA(i-1,j-1);
							COMPLETION(i,j-1)=25+DELTA(i,j-1);
							COMPLETION(i-1,j)=25+DELTA(i-1,j);
							COMPLETION(i,j)=25+DELTA(i,j);
							break;
						default:
							COMPLETION(i,j)=(*(constructionData+16*MODULE(i,j)))+DELTA(i,j);
							break;
					}
				}
			}
		}
	}

	UpdateRadarStats(baseIdx);
}
__declspec(naked) void BaseFacilityCompletedStub(int baseIdx)
{
	__asm
	{
		mov _edi,edi
		jmp BaseFacilityCompleted
	}
}

//Base build stacking
void ModBaseBuildingStacking()
{
	InsertJmp(0x433E0D,PatchBaseConstruction);
	InsertJmp(0x433CF3,PatchBaseConstructionHangar);
	InsertJmp(0x43A41C,PatchValidSquares);
	InsertCall(0x433651,BaseDrawRemainingDays);
	InsertCall(0x441AAC,BaseFacilityCompletedStub);
}

void BaseFacilityDismantleConstructionCrash()
{
	char patch[]={0x83};
	PatchInPlace(0x441ABA,patch,1);
}

/*
00 - Access Lift
01 - Living Quarters
02 - Laboratory
03 - Workshop
04 - Small Radar System
05 - Large Radar System
06 - Missile Defense
07 - General Stores
08 - Alien Containment
09 - Laser Defense
0A - Plasma Defense
0B - Fusion Ball Defense
0C - Grav Shield
0D - Mind Shield
0E - Psionic Laboratory
0F - Hyper-wave Decoder
10 - Hangar (Top Left)
11 - Hangar (Top Right)
12 - Hangar (Bottom Left)
13 - Hangar (Bottom Right)
FF - Empty
*/
static struct {
	const char *name;
	unsigned int moduleId;
} map[]=
{
	{"AccessLift",0},
	{"LivingQuarters",1},
	{"Laboratory",2},
	{"Workshop",3},
	{"SmallRadar",4},
	{"LargeRadar",5},
	{"MissileDefense",6},
	{"GeneralStores",7},
	{"AlienContainment",8},
	{"LaserDefense",9},
	{"PlasmaDefense",10},
	{"FusionBallDefense",11},
	{"GravShield",12},
	{"MindShield",13},
	{"PsionicLaboratory",14},
	{"HyperWaveDecoder",15},
	{"HangarTL",16},
	{"HangarTR",17},
	{"HangarBL",18},
	{"HangarBR",19},
	{"Empty",255},
};
static unsigned char nameToId(const char *name)
{
	for(int i=0;i<21;i++)
	{
		if(!strcmp(name,map[i].name))
		{
			return map[i].moduleId;
		}
	}
	return 0xff;
}

void SafeBaseDesign()
{
	char modules[6][6][1024];
	
	for(int i=0;i<6;i++)
	{
		char row[16];
		char line[1024];
		sprintf(row,"row%d",i+1);
		INI_READ_STRING("Initial Base",row,"",line,1024);
		sscanf(line,"%s %s %s %s %s %s",modules[i][0],modules[i][1],modules[i][2],modules[i][3],modules[i][4],modules[i][5]);
	}
	for(int i=0;i<6;i++)
	{
		for(int j=0;j<6;j++)
		{
			unsigned char id=nameToId(modules[i][j]);
			unsigned char used=id!=0xff?0:0xff;

			PatchInPlace((int)firstBaseTemplate+0x16+i*6+j,&id,1);
			PatchInPlace((int)firstBaseTemplate+0x3A+i*6+j,&used,1);
		}
	}
}
